/*
 * StreamSieve
 * A fast CUDA quadratic sieve
 * Mainly a proof of concept
 * On how GPUs are good at integer factorization
 */

#include <stdio.h>
#include <cuda.h>
#include <stdint.h>
#include <stdio.h>
#include "streamsieve.h"
#include "util.h"
#include "cuPrintf.cu"

/*
 * Max number of threads that can run on each MP
 * Currently set to 1024 for Fermi
 */
#define MAX_THREADS 1024

/*
 * Number of MP's the GPU has
 * Currently set to 16 for Fermi
 */
#define SM_COUNT 16

/*
 * The heart of the algorithm
 * Takes as input a float buffer which represents the
 * post-sieve values, a prime p and its logarithm,
 * the sieve interval, and the solution base_soln
 * to the congruence Ax^2 + Bx + C == 0 (mod p)
 * adds log(p) to the values base_soln + kp in the sieve interval
 */

__global__ void sieve(float *interval, float log_p, unsigned p,
											uint128_t min, uint128_t max, unsigned base_soln)
{
	uint128_t k_min, k_max, m_min, m_max, local_min, local_max;
	uint32_t interval_length, start_idx, end_idx, block_size;
	int n_threads, t_idx;
	//interval length better be < 4GB!
	interval_length = (max - min).to_uint32();
	n_threads = blockDim.x * gridDim.x;
	t_idx = blockIdx.x * blockDim.x + threadIdx.x;
	//Please for the love of Taylor Swift make the interval length a multiple of the thread count
	block_size = interval_length / n_threads;
	local_min = min + t_idx * block_size;
	local_max = min + (t_idx + 1) * block_size - 1;
	//want local_min < base_soln + kp < local_max
	k_min = (local_min - base_soln) / p;
	k_max = (local_max - base_soln) / p;
	m_min = p * (k_min + 1) + base_soln;
	m_max = p * (k_min + 1) + base_soln;
	start_idx = (m_min - min).to_uint32();
	end_idx = (m_max - min).to_uint32();

	for (int i = start_idx; i < end_idx; i += p) {
		interval[i] += log_p;
	}	
}

/*
 * Kernel to compute the solution to Ax^2+Bx+C == 0 (mod p)
 * over the factor base
 */
__global__ void compute_base_solns(uint512_t a, uint512_t b, uint512_t c,
																	 uint32_t* results, uint32_t* primes, uint32_t n_primes)
{
	uint32_t t_idx, start_idx, end_idx, n_threads, block_size;
	n_threads = blockDim.x * gridDim.x;
	t_idx = blockIdx.x * blockDim.x + threadIdx.x;
	//once again, Taylor Swift wants YOU to make the size of the factor base a multiple of the thread count
	block_size = n_primes / n_threads;
	start_idx = t_idx * block_size;
	end_idx = start_idx + block_size;
	for (int i = start_idx; i < end_idx; i++) {
		results[i] = quadratic(a, b, c, primes[i]);
	}	
}

int main(int argc, char** argv)
{
	printf("Welcome to StreamSieve, the CUDA Quadsieve.\n
					Nothing to see here right now, sorry!\n");
}

/*
 * NVCC is retarded
 * It doesn't have a linker
 * So we must put all of our code
 * In the same file
 */

__device__ __host__ uint128_t::uint128_t(uint32_t n)
{
	digits[0] = n;
	digits[1] = digits[2] = digits[3] = 0;
}

__device__ __host__ uint512_t::uint512_t(uint32_t n)
{
	digits[0] = n;
	#pragma unroll 15
	for (int i =  1; i < 16; i++) digits[i] = 0;
}

__device__ __host__ uint128_t::uint128_t()
{
	digits[0] = digits[1] = digits[2] = digits[3] = 0;
}

__device__ __host__ uint512_t::uint512_t()
{
	#pragma unroll 16
	for (int i = 0; i < 16; i++) digits[i] = 0;
}

__device__ __host__ uint128_t::uint128_t(uint32_t w0, uint32_t w1, uint32_t w2, uint32_t w3)
{
	digits[0] = w0;
	digits[1] = w1;
	digits[2] = w2;
	digits[3] = w3;
}

__device__ __host__ uint512_t::uint512_t(uint32_t w0, uint32_t w1, uint32_t w2, uint32_t w3,
                              uint32_t w4, uint32_t w5, uint32_t w6, uint32_t w7,
                              uint32_t w8, uint32_t w9, uint32_t w10, uint32_t w11,
                              uint32_t w12, uint32_t w13, uint32_t w14, uint32_t w15)
{
	digits[0] = w0; digits[1] = w1; digits[2] = w2; digits[3] = w3;
	digits[4] = w4; digits[5] = w5; digits[6] = w6; digits[7] = w7;
	digits[8] = w8; digits[9] = w9; digits[10] = w10; digits[11] = w11;
	digits[12] = w12; digits[13] = w13; digits[14] = w14; digits[15] = w15;
}


__device__ __host__ uint128_t uint128_t::operator+(uint128_t a)
{
  uint128_t result;
	uint32_t carry = 0;
	uint64_t temp;
	#pragma unroll 4
	for (int i = 0; i < 4; i++) {
		temp = (uint64_t) digits[i] + (uint64_t) a.digits[i] + (uint64_t) carry;
		result.digits[i] = temp & 0xFFFFFFFF;
		carry = temp >> 32;
	}
  return result;
}

__device__ __host__ uint512_t uint512_t::operator+(uint512_t a)
{
  uint512_t result;
	uint32_t carry = 0;
	uint64_t temp;
	#pragma unroll 16
	for (int i = 0; i < 16; i++) {
		temp = (uint64_t) digits[i] + (uint64_t) a.digits[i] + (uint64_t) carry;
		result.digits[i] = temp & 0xFFFFFFFF;
		carry = temp >> 32;
	}
  return result;
}

__device__ __host__ uint128_t uint128_t::operator+(unsigned a)
{
  return *this + uint128_t(a);
}

__device__ __host__ uint512_t uint512_t::operator+(unsigned a)
{
  return *this + uint512_t(a);
}

__device__ __host__ uint128_t operator+(unsigned a, uint128_t b)
{
	return b + a;
}

__device__ __host__ uint512_t operator+(unsigned a, uint512_t b)
{
	return b + a;
}

__device__ __host__ uint128_t uint128_t::operator-(uint128_t a)
{
  uint128_t result;
	uint32_t cflag, carry = 0;
	for (int i = 0; i < 4; i++) {
		cflag = (digits[i] >= carry);	
		if (cflag && digits[i] - carry >= a.digits[i]) {
			result.digits[i] = digits[i] - carry - a.digits[i];
			carry = 0;
		} else if (cflag) {
			result.digits[i] = (I32M1 - a.digits[i]) + (digits[i] - carry) + 1;
			carry = 1;
		} else {
			result.digits[i] = (I32M1 - a.digits[i]) - (carry - digits[i]) + 1;
			carry = 1;
		}
	}	 
  return result;
}

__device__ __host__ uint512_t uint512_t::operator-(uint512_t a)
{
  uint512_t result;
	uint32_t cflag, carry = 0;
	for (int i = 0; i < 16; i++) {
		cflag = (digits[i] >= carry);	
		if (cflag && digits[i] - carry >= a.digits[i]) {
			result.digits[i] = digits[i] - carry - a.digits[i];
			carry = 0;
		} else if (cflag) {
			result.digits[i] = (I32M1 - a.digits[i]) + (digits[i] - carry) + 1;
			carry = 1;
		} else {
			result.digits[i] = (I32M1 - a.digits[i]) - (carry - digits[i]) + 1;
			carry = 1;
		}
	}	 
  return result;
}

__device__ __host__ uint128_t uint128_t::operator-(unsigned a)
{
	return *this - uint128_t(a);
}

__device__ __host__ uint512_t uint512_t::operator-(unsigned a)
{
	return *this - uint512_t(a);
}

__device__ __host__ uint128_t uint128_t::operator*(uint128_t a)
{
  uint128_t result(0);
	//TODO make this function less derpy
	#pragma unroll 4
	for (int i = 0; i < 4; i++) {
		result = result + ((*this * a.digits[i]) << i);
	}
  return result;
}

__device__ __host__ uint512_t uint512_t::operator*(uint512_t a)
{
  uint512_t result(0);
	//TODO make this function less derpy
	#pragma unroll 16
	for (int i = 0; i < 16; i++) {
		result = result + ((*this * a.digits[i]) << i);
	}
  return result;
}

__device__ __host__ uint128_t uint128_t::operator*(unsigned a)
{
  uint128_t result;
	uint64_t temp;
	uint32_t carry = 0;
	#pragma unroll 4
	for (int i = 0; i < 4; i++) {
		temp = (uint64_t) digits[i] * (uint64_t) a + carry;;
		result.digits[i] = temp & 0xFFFFFFFF;
		carry = temp >> 32;	
	}
  return result;
}

__device__ __host__ uint512_t uint512_t::operator*(unsigned a)
{
  uint512_t result;
	uint64_t temp;
	uint32_t carry = 0;
	#pragma unroll 16
	for (int i = 0; i < 16; i++) {
		temp = (uint64_t) digits[i] * (uint64_t) a + carry;;
		result.digits[i] = temp & 0xFFFFFFFF;
		carry = temp >> 32;	
	}
  return result;
}

__device__ __host__ uint128_t operator*(unsigned a, uint128_t b)
{
	return b * a;
}

__device__ __host__ uint512_t operator*(unsigned a, uint512_t b)
{
	return b * a;
}

__device__ __host__ uint128_t uint128_t::operator/(unsigned a)
{
  uint128_t result;
	uint32_t last = 0;
	uint64_t temp;
	#pragma unroll 4
	for (int i = 3; i >= 0; i--) {
		temp = (((uint64_t) last) << 32) + (uint64_t) digits[i];
		result.digits[i] = temp / a;
		last = temp % a;
	}
  return result;
}

__device__ __host__ uint512_t uint512_t::operator/(unsigned a)
{
  uint512_t result;
	uint32_t last = 0;
	uint64_t temp;
	#pragma unroll 16
	for (int i = 15; i >= 0; i--) {
		temp = (((uint64_t) last) << 32) + (uint64_t) digits[i];
		result.digits[i] = temp / a;
		last = temp % a;
	}
  return result;
}

__device__ __host__ uint32_t uint128_t::operator%(unsigned a)
{
	uint32_t last = 0;
	uint64_t temp;
	#pragma unroll 4
	for (int i = 3; i >= 0; i--) {
		temp = (((uint64_t) last) << 32) + (uint64_t) digits[i];
		last = temp % a;
	}
  return last;
}

__device__ __host__ uint32_t uint512_t::operator%(unsigned a)
{
	uint32_t last = 0;
	uint64_t temp;
	#pragma unroll 16
	for (int i = 15; i >= 0; i--) {
		temp = (((uint64_t) last) << 32) + (uint64_t) digits[i];
		last = temp % a;
	}
  return last;
}

__device__ __host__ bool uint128_t::operator>(uint128_t a)
{
	#pragma unroll 4
	for (int i = 3; i >= 0; i--) {
		if (digits[i] > a.digits[i]) return true;
		if (digits[i] < a.digits[i]) return false;
	}
	return false;
}

__device__ __host__ bool uint512_t::operator>(uint512_t a)
{
	#pragma unroll 16
	for (int i = 15; i >= 0; i--) {
		if (digits[i] > a.digits[i]) return true;
		if (digits[i] < a.digits[i]) return false;
	}
	return false;
}

__device__ __host__ bool uint128_t::operator<(uint128_t a)
{
	#pragma unroll 4
	for (int i = 3; i >= 0; i--) {
		if (digits[i] > a.digits[i]) return false;
		if (digits[i] < a.digits[i]) return true;
	}
	return false;
}

__device__ __host__ bool uint512_t::operator<(uint512_t a)
{
	#pragma unroll 16
	for (int i = 15; i >= 0; i--) {
		if (digits[i] > a.digits[i]) return false;
		if (digits[i] < a.digits[i]) return true;
	}
	return false;
}

__device__ __host__ bool uint128_t::operator==(uint128_t a)
{
	#pragma unroll 4
	for (int i = 0; i < 4; i++) {
		if (digits[i] != a.digits[i]) return false;
	}
	return true;
}

__device__ __host__ bool uint512_t::operator==(uint512_t a)
{
	#pragma unroll 16
	for (int i = 0; i < 15; i++) {
		if (digits[i] != a.digits[i]) return false;
	}
	return true;
}

__device__ __host__ bool uint128_t::operator==(unsigned a)
{
	if (digits[0] != 0) return false;
	#pragma unroll 3
	for (int i = 1; i < 4; i++) {
		if (digits[i] != 0) return false;
	}
	return true;
}

__device__ __host__ bool uint512_t::operator==(unsigned a)
{
	if (digits[0] != 0) return false;
	#pragma unroll 15
	for (int i = 1; i < 16; i++) {
		if (digits[i] != 0) return false;
	}
	return true;
}

__device__ __host__ bool uint128_t::operator!=(uint128_t a)
{
	return !(*this == a);
}

__device__ __host__ bool uint512_t::operator!=(uint512_t a)
{
	return !(*this == a);
}

__device__ __host__ bool uint128_t::operator!=(unsigned a)
{
	return !(*this == a);
}

__device__ __host__ bool uint512_t::operator!=(unsigned a)
{
	return !(*this == a);
}

//TODO: this function and the next are real shitty, something needs to be done
__device__ __host__ uint128_t uint128_t::operator/(uint128_t a)
{
	uint128_t result, rem = *this;
	int a_li = a.li();
	uint32_t a_ld = a.digits[a_li];
	#pragma unroll 4
	for (int i = 3; i >= 0; i--) {
		int r_li = rem.li();
		if (r_li == -1) {
			result.digits[i] = 0;
			continue;
		}
		uint32_t qi;	
		if (rem.digits[r_li] >= a_ld) {
			qi = rem.digits[r_li] / a_ld;	
		} else if (r_li > 0) {
			uint64_t temp = ((uint64_t) rem.digits[r_li] << 32) + (uint64_t) rem.digits[r_li - 1];
			qi = (uint32_t) (temp / a_ld);
		} else {
			qi = 0;
		}
		while (qi * a << i > rem) qi--;
		digits[i] = qi;
		rem = rem - qi * a << i;
	}
	return result;
}

__device__ __host__ uint512_t uint512_t::operator/(uint512_t a)
{
	uint512_t result, rem = *this;
	int a_li = a.li();
	uint32_t a_ld = a.digits[a_li];
	#pragma unroll 15
	for (int i = 15; i >= 0; i--) {
		int r_li = rem.li();
		if (r_li == -1) {
			result.digits[i] = 0;
			continue;
		}
		uint32_t qi;	
		if (rem.digits[r_li] >= a_ld) {
			qi = rem.digits[r_li] / a_ld;	
		} else if (r_li > 0) {
			uint64_t temp = ((uint64_t) rem.digits[r_li] << 32) + (uint64_t) rem.digits[r_li - 1];
			qi = (uint32_t) (temp / a_ld);
		} else {
			qi = 0;
		}
		while (qi * a << i > rem) qi--;
		digits[i] = qi;
		rem = rem - qi * a << i;
	}
	return result;
}

//TODO these functions aren't really needed, so they're just functional, but slow stubs right now

__device__ __host__ uint128_t uint128_t::operator%(uint128_t n)
{
	return *this - n * (*this / n);
}

__device__ __host__ uint512_t uint512_t::operator%(uint512_t n)
{
	return *this - n * (*this / n);
}

__device__ __host__ uint128_t uint128_t::operator<<(uint32_t n)
{
	uint128_t result;
	#pragma unroll 4
	for (int i = 0; i < 4; i++) {
		result.digits[i] = i >= n ? digits[i - n] : 0;
	}
	return result;
}

__device__ __host__ uint512_t uint512_t::operator<<(uint32_t n)
{
	uint512_t result;
	#pragma unroll 16
	for (int i = 0; i < 16; i++) {
		result.digits[i] = i >= n ? digits[i - n] : 0;
	}
	return result;
}

__device__ __host__ uint32_t uint128_t::to_uint32()
{
  return digits[0];
}

__device__ __host__ uint32_t uint512_t::to_uint32()
{
	return digits[0];
}

__device__ __host__ int uint128_t::li()
{
	for (int i = 3; i >= 0; i--) {
		if (digits[i] != 0) return i;
	}
	return -1;
}

__device__ __host__ int uint512_t::li()
{
	for (int i = 15; i >= 0; i--) {
		if (digits[i] != 0) return i;
	}
	return -1;
}

__host__ void print_u128(uint128_t n)
{
	printf("%u %u %u %u\n", n.digits[0], n.digits[1], n.digits[2], n.digits[3]);
}

__host__ void print_u512(uint512_t n)
{
	for (int i = 0; i < 16; i++) {
		printf("%u ", n.digits[i]);
	}
	printf("\n");
}

/*
 * Begin number theory functions
 */
__device__ __host__ uint32_t modpow(uint32_t a, uint32_t n, uint32_t p)
{
	uint64_t x = 1, y = a;
	while (n > 0) {
		if (n % 2 == 1) {
			x = x * y % p;
		}
		y = y * y % p;
		n /= 2;
	}	
	return (uint32_t) x;
}

__device__ __host__ uint32_t modinv(uint512_t n, uint32_t p)
{
	uint32_t nmp = n % p;
	return modpow(nmp, p - 2, p);
}

__device__ __host__ uint32_t modsqrt(uint512_t n, uint32_t p)
{
	uint32_t nmp = n % p;
	uint32_t s, q, z, m;
	uint64_t r, t, c;
	q = p - 1;
	s = 0;
	while (q % 2 == 0) {
		s++;
		q /= 2;
	}
	for (z = 1; z < 984; z++) {
		if (modpow(z, (p - 1) / 2, p) != 1) break;
	}
	if (s == 1) return modpow(nmp, (p + 1) / 4, p);
	r = modpow(nmp, (q + 1) / 2, p);
	t = modpow(nmp, q, p);
	c = modpow(z, q, p);
	m = s;
	for (;;) {
		if (t == 1) return (uint32_t) r;
		uint64_t tt = t;
		int i;
		for (i = 0; tt != 1; i++) {
	  	tt = tt * tt % p;
		}
		uint64_t b = c;
		for (int j = 0; j < m - i - 1; j++) b = b * b % p;
		r = r  * b % p;
		t = t * b % p;
		t = t * b % p;
		c = b * b % p;
		m = i;	
	}	
}

__device__ __host__ uint32_t quadratic(uint512_t a, uint512_t b, uint512_t c, uint32_t p)
{
	uint32_t bmp = b % p;
	uint32_t cmp = c % p;
	uint64_t temp;
	//write ax^2+bx+c=a((x+b/2a)^2 + c/a - b^2/4a^3)
	uint32_t a_inv = modinv(a, p);
	temp = ((uint64_t) a_inv) << 1;
	uint32_t a2_inv = (uint32_t) (temp % p);
	temp = ((uint64_t) a_inv) << 2;
	temp = (temp * a) % p;
	temp = (temp * a) % p;
	uint32_t a43_inv = (uint32_t) temp;
	temp = (bmp * bmp) % p;
	temp = (temp * a43_inv) % p;
	uint32_t n1 = (uint32_t) temp;
	temp = (cmp * a_inv) % p;
	uint32_t n2 = (uint32_t) temp;
	uint32_t rhs;
	if (n1 >= n2) {
		rhs = n1 - n2;
	} else {
		rhs = (p - n2) + n1;
	}
	uint32_t soln = modsqrt(rhs, p);
	temp = (b * a2_inv) % p;
	uint32_t b2a = (uint32_t) temp;
	if (soln > b2a) return soln - b2a;
	return (p - b2a) + soln;
}
